@@ -1,7 +1,7 @@
 import copy
 import itertools
 from functools import partial
-from typing import Iterable, List, Tuple
+from typing import Iterable, List, Set, Tuple
 
 from isort.format import format_simplified
 
@@ -9,7 +9,7 @@ from . import parse, sorting, wrap
 from .comments import add_to_line as with_comments
 from .settings import DEFAULT_CONFIG, Config
 
-STATEMENT_DECLERATIONS: Tuple[str, ...] = ("def ", "cdef ", "cpdef ", "class ", "@", "async def")
+STATEMENT_DECLARATIONS: Tuple[str, ...] = ("def ", "cdef ", "cpdef ", "class ", "@", "async def")
 
 
 def sorted_imports(
@@ -29,7 +29,6 @@ def sorted_imports(
     formatted_output: List[str] = parsed.lines_without_imports.copy()
     remove_imports = [format_simplified(removal) for removal in config.remove_imports]
 
-    sort_ignore_case = config.force_alphabetical_sort_within_sections
     sections: Iterable[str] = itertools.chain(parsed.sections, config.forced_separate)
 
     if config.no_sections:
@@ -46,62 +45,35 @@ def sorted_imports(
         sections = base_sections + ("no_sections",)
 
     output: List[str] = []
+    seen_headings: Set[str] = set()
     pending_lines_before = False
     for section in sections:
         straight_modules = parsed.imports[section]["straight"]
         straight_modules = sorting.naturally(
-            straight_modules, key=lambda key: sorting.module_key(key, config, section_name=section)
+            straight_modules,
+            key=lambda key: sorting.module_key(
+                key, config, section_name=section, straight_import=True
+            ),
         )
         from_modules = parsed.imports[section]["from"]
         from_modules = sorting.naturally(
             from_modules, key=lambda key: sorting.module_key(key, config, section_name=section)
         )
 
-        section_output: List[str] = []
+        straight_imports = _with_straight_imports(
+            parsed, config, straight_modules, section, remove_imports, import_type
+        )
+        from_imports = _with_from_imports(
+            parsed, config, from_modules, section, remove_imports, import_type
+        )
+
+        lines_between = [""] * (
+            config.lines_between_types if from_modules and straight_modules else 0
+        )
         if config.from_first:
-            section_output = _with_from_imports(
-                parsed,
-                config,
-                from_modules,
-                section,
-                section_output,
-                sort_ignore_case,
-                remove_imports,
-                import_type,
-            )
-            if config.lines_between_types and from_modules and straight_modules:
-                section_output.extend([""] * config.lines_between_types)
-            section_output = _with_straight_imports(
-                parsed,
-                config,
-                straight_modules,
-                section,
-                section_output,
-                remove_imports,
-                import_type,
-            )
+            section_output = from_imports + lines_between + straight_imports
         else:
-            section_output = _with_straight_imports(
-                parsed,
-                config,
-                straight_modules,
-                section,
-                section_output,
-                remove_imports,
-                import_type,
-            )
-            if config.lines_between_types and from_modules and straight_modules:
-                section_output.extend([""] * config.lines_between_types)
-            section_output = _with_from_imports(
-                parsed,
-                config,
-                from_modules,
-                section,
-                section_output,
-                sort_ignore_case,
-                remove_imports,
-                import_type,
-            )
+            section_output = straight_imports + lines_between + from_imports
 
         if config.force_sort_within_sections:
             # collapse comments
@@ -126,6 +98,7 @@ def sorted_imports(
                     force_to_top=config.force_to_top,
                     lexicographical=config.lexicographical,
                     length_sort=config.length_sort,
+                    reverse_relative=config.reverse_relative,
                 ),
             )
 
@@ -134,12 +107,6 @@ def sorted_imports(
             for line in new_section_output:
                 comments = getattr(line, "comments", ())
                 if comments:
-                    if (
-                        config.ensure_newline_before_comments
-                        and section_output
-                        and section_output[-1]
-                    ):
-                        section_output.append("")
                     section_output.extend(comments)
                 section_output.append(str(line))
 
@@ -152,7 +119,9 @@ def sorted_imports(
                 continue
 
             section_title = config.import_headings.get(section_name.lower(), "")
-            if section_title:
+            if section_title and section_title not in seen_headings:
+                if config.dedup_headings:
+                    seen_headings.add(section_title)
                 section_comment = f"# {section_title}"
                 if section_comment not in parsed.lines_without_imports[0:1]:
                     section_output.insert(0, section_comment)
@@ -166,6 +135,9 @@ def sorted_imports(
         else:
             pending_lines_before = pending_lines_before or not no_lines_before
 
+    if config.ensure_newline_before_comments:
+        output = _ensure_newline_before_comment(output)
+
     while output and output[-1].strip() == "":
         output.pop()  # pragma: no cover
     while output and output[0].strip() == "":
@@ -214,7 +186,7 @@ def sorted_imports(
 
         if config.lines_after_imports != -1:
             formatted_output[imports_tail:0] = ["" for line in range(config.lines_after_imports)]
-        elif extension != "pyi" and next_construct.startswith(STATEMENT_DECLERATIONS):
+        elif extension != "pyi" and next_construct.startswith(STATEMENT_DECLARATIONS):
             formatted_output[imports_tail:0] = ["", ""]
         else:
             formatted_output[imports_tail:0] = [""]
@@ -225,7 +197,10 @@ def sorted_imports(
             new_out_lines.append(line)
             if line in parsed.import_placements:
                 new_out_lines.extend(parsed.place_imports[parsed.import_placements[line]])
-                if len(formatted_output) <= index or formatted_output[index + 1].strip() != "":
+                if (
+                    len(formatted_output) <= (index + 1)
+                    or formatted_output[index + 1].strip() != ""
+                ):
                     new_out_lines.append("")
         formatted_output = new_out_lines
 
@@ -237,12 +212,10 @@ def _with_from_imports(
     config: Config,
     from_modules: Iterable[str],
     section: str,
-    section_output: List[str],
-    ignore_case: bool,
     remove_imports: List[str],
     import_type: str,
 ) -> List[str]:
-    new_section_output = section_output.copy()
+    output: List[str] = []
     for module in from_modules:
         if module in remove_imports:
             continue
@@ -252,10 +225,15 @@ def _with_from_imports(
         if not config.no_inline_sort or (
             config.force_single_line and module not in config.single_line_exclusions
         ):
+            ignore_case = config.force_alphabetical_sort_within_sections
             from_imports = sorting.naturally(
                 from_imports,
                 key=lambda key: sorting.module_key(
-                    key, config, True, ignore_case, section_name=section
+                    key,
+                    config,
+                    True,
+                    ignore_case,
+                    section_name=section,
                 ),
             )
         if remove_imports:
@@ -283,18 +261,18 @@ def _with_from_imports(
                     else:
                         from_imports[idx : (idx + 1)] = as_imports.pop(from_import)
 
+        only_show_as_imports = False
+        comments = parsed.categorized_comments["from"].pop(module, ())
+        above_comments = parsed.categorized_comments["above"]["from"].pop(module, None)
         while from_imports:
-            comments = parsed.categorized_comments["from"].pop(module, ())
-            above_comments = parsed.categorized_comments["above"]["from"].pop(module, None)
             if above_comments:
-                if new_section_output and config.ensure_newline_before_comments:
-                    new_section_output.append("")
-                new_section_output.extend(above_comments)
+                output.extend(above_comments)
+                above_comments = None
 
             if "*" in from_imports and config.combine_star:
                 import_statement = wrap.line(
                     with_comments(
-                        comments,
+                        _with_star_comments(parsed, module, list(comments or ())),
                         f"{import_start}*",
                         removed=config.ignore_comments,
                         comment_prefix=config.comment_prefix,
@@ -302,7 +280,10 @@ def _with_from_imports(
                     parsed.line_separator,
                     config,
                 )
-                from_imports = []
+                from_imports = [
+                    from_import for from_import in from_imports if from_import in as_imports
+                ]
+                only_show_as_imports = True
             elif config.force_single_line and module not in config.single_line_exclusions:
                 import_statement = ""
                 while from_imports:
@@ -321,14 +302,17 @@ def _with_from_imports(
                             f"{comments and ';' or config.comment_prefix} " f"{comment}"
                         )
                     if from_import in as_imports:
-                        if parsed.imports[section]["from"][module][from_import]:
-                            new_section_output.append(
+                        if (
+                            parsed.imports[section]["from"][module][from_import]
+                            and not only_show_as_imports
+                        ):
+                            output.append(
                                 wrap.line(single_import_line, parsed.line_separator, config)
                             )
                         from_comments = parsed.categorized_comments["straight"].get(
                             f"{module}.{from_import}"
                         )
-                        new_section_output.extend(
+                        output.extend(
                             with_comments(
                                 from_comments,
                                 wrap.line(import_start + as_import, parsed.line_separator, config),
@@ -338,19 +322,27 @@ def _with_from_imports(
                             for as_import in sorting.naturally(as_imports[from_import])
                         )
                     else:
-                        new_section_output.append(
-                            wrap.line(single_import_line, parsed.line_separator, config)
-                        )
+                        output.append(wrap.line(single_import_line, parsed.line_separator, config))
                     comments = None
             else:
                 while from_imports and from_imports[0] in as_imports:
                     from_import = from_imports.pop(0)
                     as_imports[from_import] = sorting.naturally(as_imports[from_import])
-                    from_comments = parsed.categorized_comments["straight"].get(
-                        f"{module}.{from_import}"
+                    from_comments = (
+                        parsed.categorized_comments["straight"].get(f"{module}.{from_import}") or []
                     )
-                    if parsed.imports[section]["from"][module][from_import]:
-                        new_section_output.append(
+                    if (
+                        parsed.imports[section]["from"][module][from_import]
+                        and not only_show_as_imports
+                    ):
+                        specific_comment = (
+                            parsed.categorized_comments["nested"]
+                            .get(module, {})
+                            .pop(from_import, None)
+                        )
+                        if specific_comment:
+                            from_comments.append(specific_comment)
+                        output.append(
                             wrap.line(
                                 with_comments(
                                     from_comments,
@@ -362,32 +354,42 @@ def _with_from_imports(
                                 config,
                             )
                         )
-                    new_section_output.extend(
-                        wrap.line(
-                            with_comments(
-                                from_comments,
-                                import_start + as_import,
-                                removed=config.ignore_comments,
-                                comment_prefix=config.comment_prefix,
-                            ),
-                            parsed.line_separator,
-                            config,
+                        from_comments = []
+
+                    for as_import in as_imports[from_import]:
+                        specific_comment = (
+                            parsed.categorized_comments["nested"]
+                            .get(module, {})
+                            .pop(as_import, None)
                         )
-                        for as_import in as_imports[from_import]
-                    )
+                        if specific_comment:
+                            from_comments.append(specific_comment)
+
+                        output.append(
+                            wrap.line(
+                                with_comments(
+                                    from_comments,
+                                    import_start + as_import,
+                                    removed=config.ignore_comments,
+                                    comment_prefix=config.comment_prefix,
+                                ),
+                                parsed.line_separator,
+                                config,
+                            )
+                        )
+
+                        from_comments = []
 
-                star_import = False
                 if "*" in from_imports:
-                    new_section_output.append(
+                    output.append(
                         with_comments(
-                            comments,
+                            _with_star_comments(parsed, module, list(comments or ())),
                             f"{import_start}*",
                             removed=config.ignore_comments,
                             comment_prefix=config.comment_prefix,
                         )
                     )
                     from_imports.remove("*")
-                    star_import = True
                     comments = None
 
                 for from_import in copy.copy(from_imports):
@@ -404,9 +406,7 @@ def _with_from_imports(
                         single_import_line += (
                             f"{comments and ';' or config.comment_prefix} " f"{comment}"
                         )
-                        new_section_output.append(
-                            wrap.line(single_import_line, parsed.line_separator, config)
-                        )
+                        output.append(wrap.line(single_import_line, parsed.line_separator, config))
                         from_imports.remove(from_import)
                         comments = None
 
@@ -419,15 +419,16 @@ def _with_from_imports(
                     )
                 ):
                     from_import_section.append(from_imports.pop(0))
-                if star_import:
-                    import_statement = import_start + (", ").join(from_import_section)
-                else:
-                    import_statement = with_comments(
-                        comments,
-                        import_start + (", ").join(from_import_section),
-                        removed=config.ignore_comments,
-                        comment_prefix=config.comment_prefix,
+                if config.combine_as_imports:
+                    comments = (comments or []) + list(
+                        parsed.categorized_comments["from"].pop(f"{module}.__combined_as__", ())
                     )
+                import_statement = with_comments(
+                    comments,
+                    import_start + (", ").join(from_import_section),
+                    removed=config.ignore_comments,
+                    comment_prefix=config.comment_prefix,
+                )
                 if not from_import_section:
                     import_statement = ""
 
@@ -473,8 +474,8 @@ def _with_from_imports(
                     import_statement = wrap.line(import_statement, parsed.line_separator, config)
 
             if import_statement:
-                new_section_output.append(import_statement)
-    return new_section_output
+                output.append(import_statement)
+    return output
 
 
 def _with_straight_imports(
@@ -482,11 +483,10 @@ def _with_straight_imports(
     config: Config,
     straight_modules: Iterable[str],
     section: str,
-    section_output: List[str],
     remove_imports: List[str],
     import_type: str,
 ) -> List[str]:
-    new_section_output = section_output.copy()
+    output: List[str] = []
     for module in straight_modules:
         if module in remove_imports:
             continue
@@ -504,10 +504,8 @@ def _with_straight_imports(
 
         comments_above = parsed.categorized_comments["above"]["straight"].pop(module, None)
         if comments_above:
-            if new_section_output and config.ensure_newline_before_comments:
-                new_section_output.append("")
-            new_section_output.extend(comments_above)
-        new_section_output.extend(
+            output.extend(comments_above)
+        output.extend(
             with_comments(
                 parsed.categorized_comments["straight"].get(module),
                 idef,
@@ -517,7 +515,7 @@ def _with_straight_imports(
             for idef in import_definition
         )
 
-    return new_section_output
+    return output
 
 
 def _output_as_string(lines: List[str], line_separator: str) -> str:
@@ -537,3 +535,24 @@ class _LineWithComments(str):
         instance = super().__new__(cls, value)  # type: ignore
         instance.comments = comments
         return instance
+
+
+def _ensure_newline_before_comment(output):
+    new_output: List[str] = []
+
+    def is_comment(line):
+        return line and line.startswith("#")
+
+    for line, prev_line in zip(output, [None] + output):
+        if is_comment(line) and prev_line != "" and not is_comment(prev_line):
+            new_output.append("")
+        new_output.append(line)
+    return new_output
+
+
+def _with_star_comments(parsed: parse.ParsedContent, module: str, comments: List[str]) -> List[str]:
+    star_comment = parsed.categorized_comments["nested"].get(module, {}).pop("*", None)
+    if star_comment:
+        return comments + [star_comment]
+    else:
+        return comments
